CUSTOM_TABLE_FIELDS_ON_URL = """&c=0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,17,18,19,20,21,23,22,78,127,128,24,25,85,26,27,28,29,30,31,32,33,34,35,36,37,
38,39,40,41,42,43,44,45,46,47,48,49,50,51,52,53,54,57,58,59,62,63,64,65,66,67,68,69,73,74,76,77,79,80,81,82,83,84,120,130,131,132,133,134"""

FINVIZ_DROP_COLUMNS = ["Earnings", "52W Range", "Return% 1Y", "Index", "52W High", "52W Low"]


CUSTOM_TABLE_ALL_FIELDS_NEW = [
    'Ticker',
    'Company',
    'Sector',
    'Industry',
    'Country',
    'Market Cap',
    'P/E',
    'Fwd P/E',
    'PEG',
    'P/S',
    'P/B',
    'P/C',
    'P/FCF',
    'Dividend',
    'Payout Ratio',
    'EPS',
    'EPS next Q',
    'EPS This Y',
    'EPS Next Y',
    'EPS Past 5Y',
    'EPS Next 5Y',
    'Sales Past 5Y',
    'Sales Q/Q',
    'Sales YoY TTM',
    'EPS Q/Q',
    'Sales',
    'Income',
    'EPS Surprise',
    'Revenue Surprise',
    'Outstanding',
    'Float',
    'Float %',
    'Insider Own',
    'Insider Trans',
    'Inst Own',
    'Inst Trans',
    'Short Float',
    'Short Ratio',
    'Short Interest',
    'ROA',
    'ROE',
    'ROIC',
    'Curr R',
    'Quick R',
    'LTDebt/Eq',
    'Debt/Eq',
    'Gross M',
    'Oper M',
    'Profit M',
    'Perf Week',
    'Perf Month',
    'Perf Quart',
    'Perf Half',
    'Perf Year',
    'Perf YTD',
    'Beta',
    'ATR',
    'Volatility W',
    'Volatility M',
    'SMA20',
    'SMA50',
    'SMA200',
    '52W High',
    '52W Low',
    'RSI',
    'Earnings',
    'Target Price',
    'Book/sh',
    'Cash/sh',
    'Employees',
    'Index',
    'Optionable',
    'Prev Close',
    'Shortable',
    'Recom',
    'Avg Volume',
    'Rel Volume',
    'Volume',
    'Price',
    'Change',
    'Return% 1Y',
    'Dividend TTM',
    'Dividend Ex Date',
    'EPS YoY TTM',
    '52W Range'
]

POST_GENERATED_FIELDS = [
    '52W_NORM',
    'LTDEBT/EQ',
    'DIVIDEND_YIELD_EST',
    'DIVIDEND_YIELD_TTM',
    'IS_S&P500',
    'IS_RUSSELL',
    'IS_NASDAQ',
    'IS_AMC',
    'IS_BMO',
    'DIVIDEND_EX-DATE',
    'IS_OPTIONABLE',
    'IS_SHORTABLE',
    'VOLATILITY_RANGE',
    'IS_USA',
    'DATADATE',
]


MONEY_COLUMNS = [
    'Avg Volume', 
    'Outstanding',
    'Market Cap',
    'Float',
    'Income',
    'Sales',
    'Short Interest',
    'Outstanding',
    'Float'
]

PERCENTAJE_COLUMNS = [
    'Perf Week', 
    'Perf Month', 
    'Perf Quart', 
    'Perf Year', 
    'Perf YTD', 
    'Volatility W', 
    'Volatility M', 
    'Change', 
    'Insider Own',
    'EPS Next Y',
    'EPS This Y',
    'EPS Past 5Y',
    'EPS Next 5Y',
    'Sales Past 5Y',
    'Insider Trans',
    'Inst Own',
    'Perf Quart',
    'Inst Trans',
    'Perf Half',
    'ROA',
    'ROE',
    'ROIC',
    '52W High',
    'Gross M',
    '52W Low',
    'Sales Q/Q',
    'Oper M',
    'EPS Q/Q',
    'Profit M',
    'Payout Ratio',
    'SMA20',
    'SMA50',
    'SMA200',
    'Dividend',
    'Payout Ratio',
    'EPS Surprise',
    'Revenue Surprise',
    'Float %',
    'Oper M',
    'Profit M',
    'Short Float',
    'EPS YoY TTM',
    'Sales YoY TTM',
]

NUMERIC_COLUMNS = [
    'Dividend TTM',
    'P/FCF',
    'Beta',
    'EPS YoY TTM',
    'ATR',
    'Employees',
    'Curr R',
    'Quick R',
    'RSI',
    'Debt/Eq',
    'LTDebt/Eq',
    'Recom',
    'Rel Volume',
    'Price',
    'Volume',
    'P/E',
    'Fwd P/E',
    'PEG',
    'P/S',
    'Book/sh',
    'P/B',
    'Target Price',
    'Cash/sh',
    'P/C',
    'EPS next Q',
    'Fwd P/E',
    'EPS',
    'Short Ratio',
    'Short Interest',
    'RSI',
    'Avg Volume',
    'Return% 1Y',
    'Outstanding',
    'Float',
    'ATR',
    'Prev Close',
]

BOOLEAN_COLUMNS = [
    'Optionable',
    'Shortable'
]

TICKER_KEY = 'Ticker'
VALID_SCOPES_EARNING_SCOPES = [
    'last_week', 
    'this_week', 
    'next_week', 
    'today_bmo', 
    'yesterday_amc', 
    'today',
    'this_month'
]



from earningspy.generators.yahoo.time_series import get_portfolio
from earningspy.generators.yahoo.async_timeseries import get_portfolio as async_get_portfolio

TICKERS = ['AAPL', 'MSFT', 'GOOGL', 'AMZN']

def test_async_timeseries():
    """Test asynchronous time series data retrieval has low NaN ratio."""
    data = async_get_portfolio(assets=TICKERS, from_='3m')

    assert data is not None, "Data is None"
    missing_tickers = [t for t in TICKERS if t not in data.columns]
    assert not missing_tickers, f"Missing tickers: {missing_tickers}"

    for ticker in TICKERS:
        nan_ratio = data[ticker].isna().mean()
        assert nan_ratio <= 0.95, f"{ticker} has too many NaNs: {nan_ratio:.2%}"


def test_time_series():
    """Test synchronous timeseries data retrieval."""
    data = get_portfolio(assets=TICKERS, from_='3m')

    assert data is not None, "Data is None"
    missing_tickers = [t for t in TICKERS if t not in data.columns]
    assert not missing_tickers, f"Missing tickers: {missing_tickers}"

    for ticker in TICKERS:
        nan_ratio = data[ticker].isna().mean()
        assert nan_ratio <= 0.95, f"{ticker} has too many NaNs: {nan_ratio:.2%}"



TICKER_KEY = 'Ticker'
TICKER_KEY_CAPITAL = 'TICKER'
COMPANY_KEY_CAPITAL = 'COMPANY'
DAYS_TO_EARNINGS_KEY_CAPITAL='DAYS_LEFT'
DAYS_TO_EARNINGS_KEY_BEFORE_FORMAT='days_left'

MARKET_DATA_TICKERS = ['^GSPC', '^TNX', '^RUT', '^VIX']
TBILL_10_YEAR = '^TNX'

FINVIZ_EARNINGS_DATE_KEY = 'EARNINGS_DATE'
DAYS_LEFT_KEY = 'DAYS_LEFT'
ALLOWED_CAPITALIZATIONS = ['micro', 'small', 'medium']

# Post processing fields 
ABS_RET_KEY='1+{}_RET'
EXP_RET_KEY='EXP.RET_{}'
BETA_KEY='BETA'
RF_KEY='1+{}_RF'
SP_500_TICKER='^GSPC'
MARK_EXP_KEY='MARK_EXP_{}'
CAPM_KEY='CAPM_{}'
TBILL_10_YEAR='^TNX'
VIX_TICKER='^VIX'
VIX_KEY='1+{}_VIX'
EARNING_VIX_KEY='EARNING_VIX'
CAR_KEY='CAR_{}'
BHAR_KEY='BHAR_{}'
DATADATE_KEY='DATADATE'
ALLOWED_WINDOWS = [3, 30, 60]

AVAILABLE_CHECK_COLUMNS = [
    EXP_RET_KEY,
    ABS_RET_KEY,
    MARK_EXP_KEY,
    CAPM_KEY,
    CAR_KEY,
    BHAR_KEY,
]

FIELDS_ORDER = [
'DAYS_LEFT',
'TICKER',
'COMPANY',
'PRICE',
'52W_NORM',
'AVG_VOLUME',
'DIVIDEND_EX_DATE',
'DIVIDEND',
'PAYOUT_RATIO',
'SECTOR',
'INDUSTRY',
'COUNTRY',
'MARKET_CAP',
'P/E',
'FWD_P/E',
'PEG',
'P/S',
'P/B',
'P/C',
'P/FCF',
'EPS',
'EPS_NEXT_Q',
'EPS_THIS_Y',
'EPS_NEXT_Y',
'EPS_PAST_5Y',
'EPS_NEXT_5Y',
'SALES_PAST_5Y',
'SALES_Q/Q',
'SALES_YOY_TTM',
'EPS_Q/Q',
'SALES',
'INCOME',
'EPS_SURPRISE',
'REVENUE_SURPRISE',
'OUTSTANDING',
'VOLUME',
'REL_VOLUME',
'FLOAT',
'FLOAT_%',
'INSIDER_OWN',
'INSIDER_TRANS',
'INST_OWN',
'INST_TRANS',
'SHORT_FLOAT',
'SHORT_RATIO',
'SHORT_INTEREST',
'ROA',
'ROE',
'ROIC',
'CURR_R',
'QUICK_R',
'LTDEBT/EQ',
'DEBT/EQ',
'GROSS_M',
'OPER_M',
'PROFIT_M',
'PERF_WEEK',
'PERF_MONTH',
'PERF_QUART',
'PERF_HALF',
'PERF_YEAR',
'PERF_YTD',
'BETA',
'ATR',
'VOLATILITY_W',
'VOLATILITY_M',
'SMA20',
'SMA50',
'SMA200',
'RSI',
'TARGET_PRICE',
'BOOK/SH',
'CASH/SH',
'EMPLOYEES',
'OPTIONABLE',
'PREV_CLOSE',
'SHORTABLE',
'RECOM',
'CHANGE',
'DIVIDEND_TTM',
'EPS_YOY_TTM',
'IS_S&P500',
'IS_RUSSELL',
'IS_NASDAQ',
'IS_DOW_JONES',
'IS_AMC',
'IS_BMO',
'IS_USA',
'DATADATE'
]


import pytest
from earningspy.generators.finviz.data import (
    _get_screener_data, 
    get_by_earnings_date, 
    get_by_tickers,
)

EXPECTED_COLUMNS = [
    'Ticker', 'Company', 'Sector', 'Industry', 'Country', 'Market Cap',
    'P/E', 'Fwd P/E', 'PEG', 'P/S', 'P/B', 'P/C', 'P/FCF', 'Dividend',
    'Payout Ratio', 'EPS', 'EPS next Q', 'EPS This Y', 'EPS Next Y',
    'EPS Past 5Y', 'EPS Next 5Y', 'Sales Past 5Y', 'Sales Q/Q',
    'Sales YoY TTM', 'EPS Q/Q', 'Sales', 'Income', 'EPS Surprise',
    'Revenue Surprise', 'Outstanding', 'Float', 'Float %', 'Insider Own',
    'Insider Trans', 'Inst Own', 'Inst Trans', 'Short Float', 'Short Ratio',
    'Short Interest', 'ROA', 'ROE', 'ROIC', 'Curr R', 'Quick R',
    'LTDebt/Eq', 'Debt/Eq', 'Gross M', 'Oper M', 'Profit M', 'Perf Week',
    'Perf Month', 'Perf Quart', 'Perf Half', 'Perf Year', 'Perf YTD',
    'Beta', 'ATR', 'Volatility W', 'Volatility M', 'SMA20', 'SMA50',
    'SMA200', 'RSI', 'Target Price', 'Book/sh', 'Cash/sh', 'Employees',
    'Optionable', 'Prev Close', 'Shortable', 'Recom', 'Avg Volume',
    'Rel Volume', 'Volume', 'Price', 'Change', 'Dividend TTM',
    'Dividend Ex Date', 'EPS YoY TTM', '52W_NORM', 'IS_S&P500',
    'IS_RUSSELL', 'IS_NASDAQ', 'IS_DOW_JONES', 'IS_AMC', 'IS_BMO', 'IS_USA',
    'EARNINGS_DATE', 'DATADATE'
]

EXCLUDE_NAN_CHECK = [
    '52W_NORM', 'IS_S&P500', 'IS_RUSSELL', 'IS_NASDAQ', 'IS_DOW_JONES',
    'IS_AMC', 'IS_BMO', 'IS_USA', 'EARNINGS_DATE', 'DATADATE', 'Dividend Ex Date'
]


@pytest.mark.parametrize("filters", [
    # industries with few stocks to avoid running for too long time
    # 'ind_gambling',
    # 'ind_semiconductors',
    'ind_solar'
])
def test_get_screener_data(filters):
    print(f"Testing filter: {filters}")
    data = _get_screener_data(filters=filters, order='marketcap')
    nan_columns = [col for col in data.columns if col not in EXCLUDE_NAN_CHECK and data[col].isnull().any()]

    assert data.columns.tolist() == EXPECTED_COLUMNS, "DataFrame columns do not match expected columns"
    assert not data.empty, "DataFrame should not be empty"
    assert not nan_columns, f"DataFrame contains NaNs in columns: {nan_columns}"


@pytest.mark.parametrize("scopes", [
    # 'last_week', 
    # 'this_week', 
    'next_week', 
])
def test_get_by_earnings_date(scopes):
    """Test the get_by_earnings_date function."""
    print(f"Testing scope: {scopes}")
    data = get_by_earnings_date(scopes)
    nan_columns = [col for col in data.columns if col not in EXCLUDE_NAN_CHECK and data[col].isnull().any()]

    assert data.columns.tolist() == EXPECTED_COLUMNS, "DataFrame columns do not match expected columns"
    assert not data.empty, "DataFrame should not be empty"
    assert not nan_columns, f"DataFrame contains NaNs in columns: {nan_columns}"

def test_get_by_tickers():
    """Test the get_by_tickers function."""
    tickers = ['AAPL', 'MSFT', 'GOOGL']
    data = get_by_tickers(tickers)
    nan_columns = [col for col in data.columns if col not in EXCLUDE_NAN_CHECK and data[col].isnull().any()]

    assert data.columns.tolist() == EXPECTED_COLUMNS, "DataFrame columns do not match expected columns"
    assert not data.empty, "DataFrame should not be empty"
    assert not nan_columns, f"DataFrame contains NaNs in columns: {nan_columns}"



import csv

import requests
from lxml import html
from user_agent import generate_user_agent

from earningspy.generators.finviz.helper_functions.display_functions import create_table_string
from earningspy.generators.finviz.helper_functions.error_handling import (InvalidPortfolioID,
                                                    InvalidTicker,
                                                    NonexistentPortfolioName)
from earningspy.generators.finviz.helper_functions.request_functions import http_request_get
from earningspy.generators.finviz.helper_functions.scraper_functions import get_table

LOGIN_URL = "https://finviz.com/login_submit.ashx"
PRICE_REQUEST_URL = "https://finviz.com/request_quote.ashx"
PORTFOLIO_URL = "https://finviz.com/portfolio.ashx"
PORTFOLIO_SUBMIT_URL = "https://finviz.com/portfolio_submit.ashx"
PORTFOLIO_DIGIT_COUNT = 9  # Portfolio ID is always 9 digits
PORTFOLIO_HEADERS = [
    "No.",
    "Ticker",
    "Company",
    "Price",
    "Change%",
    "Volume",
    "Transaction",
    "Date",
    "Shares",
    "Cost",
    "Market Value",
    "Gain$",
    "Gain%",
    "Change$",
]


class Portfolio(object):
    """ Used to interact with FinViz Portfolio. """

    def __init__(self, email, password, portfolio=None):
        """
        Logs in to FinViz and send a GET request to the portfolio.
        """

        payload = {"email": email, "password": password}

        # Create a session and log in by sending a POST request
        self._session = requests.session()
        auth_response = self._session.post(
            LOGIN_URL, data=payload, headers={"User-Agent": generate_user_agent()}
        )

        if not auth_response.ok:  # If the post request wasn't successful
            auth_response.raise_for_status()

        # Get the parsed HTML and the URL of the base portfolio page
        self._page_content, self.portfolio_url = http_request_get(
            url=PORTFOLIO_URL, session=self._session, parse=False
        )

        # If the user has not created a portfolio it redirects the request to <url>?v=2)
        self.created = True
        if self.portfolio_url == f"{PORTFOLIO_URL}?v=2":
            self.created = False

        if self.created:
            if portfolio:
                self._page_content, _ = self.__get_portfolio_url(portfolio)

            self.data = get_table(self._page_content, PORTFOLIO_HEADERS)

    def __str__(self):
        """ Returns a readable representation of a table. """

        table_list = [PORTFOLIO_HEADERS]

        for row in self.data:
            table_list.append([row[col] or "" for col in PORTFOLIO_HEADERS])

        return create_table_string(table_list)

    def create_portfolio(self, name, file, drop_invalid_ticker=False):
        """
        Creates a new portfolio from a .csv file.

        The .csv file must be in the following format:
        Ticker,Transaction,Date,Shares,Price
        NVDA,2,14-04-2018,43,148.26
        AAPL,1,01-05-2019,12
        WMT,1,25-02-2015,20
        ENGH:CA,1,,1,

        (!) For transaction - 1 = BUY, 2 = SELL
        (!) Note that if the price is omitted the function will take today's ticker price
        """

        data = {
            "portfolio_id": "0",
            "portfolio_name": name,
        }

        with open(file, "r") as infile:
            reader = csv.reader(infile)
            next(reader, None)  # Skip the headers

            for row_number, row in enumerate(reader, 0):
                row_number_string = str(row_number)
                data["ticker" + row_number_string] = row[0]
                data["transaction" + row_number_string] = row[1]
                data["date" + row_number_string] = row[2]
                data["shares" + row_number_string] = row[3]

                try:
                    # empty string is no price, so try get today's price
                    assert data["price" + row_number_string] != ""
                    data["price" + row_number_string] = row[4]
                except (IndexError, KeyError):
                    current_price_page, _ = http_request_get(
                        PRICE_REQUEST_URL, payload={"t": row[0]}, parse=True
                    )

                    # if price not available on finviz don't upload that ticker to portfolio
                    if current_price_page.text == "NA":
                        if not drop_invalid_ticker:
                            raise InvalidTicker(row[0])
                        del data["ticker" + row_number_string]
                        del data["transaction" + row_number_string]
                        del data["date" + row_number_string]
                        del data["shares" + row_number_string]
                    else:
                        data["price" + row_number_string] = current_price_page.text
        self._session.post(PORTFOLIO_SUBMIT_URL, data=data)

    def __get_portfolio_url(self, portfolio_name):
        """ Private function used to return the portfolio url from a given id/name. """

        # If the user has provided an ID (Portfolio ID is always an int)
        if isinstance(portfolio_name, int):
            # Raise error for invalid portfolio ID
            if not len(str(portfolio_name)) == PORTFOLIO_DIGIT_COUNT:
                raise InvalidPortfolioID(portfolio_name)
            else:
                return http_request_get(
                    url=f"{PORTFOLIO_URL}?pid={portfolio_name}",
                    session=self._session,
                    parse=False,
                )
        else:  # else the user has passed a name
            # We remove the first element, since it's redundant
            for portfolio in html.fromstring(self._page_content).cssselect("option")[
                1:
            ]:
                if portfolio.text == portfolio_name:
                    return http_request_get(
                        url=f"{PORTFOLIO_URL}?pid={portfolio.get('value')}",
                        session=self._session,
                        parse=False,
                    )
            # Raise Non-existing PortfolioName if none of the names match
            raise NonexistentPortfolioName(portfolio_name)



from earningspy.generators.finviz.config import connection_settings


class NoResults(Exception):
    """ Raise when there are no results found. """

    def __init__(self, query):
        super(NoResults, self).__init__(f"No results found for query: {query}")


class InvalidTableType(Exception):
    """ Raise when the given table type is invalid. """

    def __init__(self, arg):
        super(InvalidTableType, self).__init__(f"Invalid table type called: {arg}")


class TooManyRequests(Exception):
    """ Raise when HTTP request fails because too many requests were sent to FinViz at once. """

    def __init__(self, arg):
        super(TooManyRequests, self).__init__(f"Too many HTTP requests at once: {arg}")


class InvalidPortfolioID(Exception):
    """ Raise when the given portfolio id is invalid. """

    def __int__(self, portfolio_id):
        super(InvalidPortfolioID, self).__init__(
            f"Invalid portfolio with ID: {portfolio_id}"
        )


class NonexistentPortfolioName(Exception):
    """ Raise when the given portfolio name is nonexistent. """

    def __init__(self, name):
        super(NonexistentPortfolioName, self).__init__(
            f"Nonexistent portfolio with name: {name}"
        )


class NoPortfolio(Exception):
    """ Raise when the user has not created a portfolio. """

    def __int__(self, func_name):
        super(NoPortfolio, self).__init__(
            "Function ({func_name}) cannot be called because "
            "there is no existing portfolio."
        )


class InvalidTicker(Exception):
    """ Raise when the given ticker is nonexistent or unavailable on FinViz.  """

    def __init__(self, ticker):
        super(InvalidTicker, self).__init__(
            f"Unable to find {ticker} since it is non-existent or unavailable on FinViz."
        )


class ConnectionTimeout(Exception):
    """ The request has timed out while trying to connect to the remote server. """

    def __init__(self, webpage_link):
        super(ConnectionTimeout, self).__init__(
            f'Connection timed out after {connection_settings["CONNECTION_TIMEOUT"]} while trying to reach {webpage_link}'
        )



from time import sleep
import requests
from datetime import datetime as dt
from dateutil.relativedelta import relativedelta
import pandas as pd
from tqdm import tqdm

# Fixing this problem: https://www.reddit.com/r/sheets/comments/1farvxr/broken_yahoo_finance_url/


def get_range(range_from, end_date):

    accepted_values = ['3m', '9m', '1y', '5y', '10y']
    if range_from not in accepted_values:
        raise Exception('Invalid from value')

    if range_from == '3m':
        start_date = end_date - relativedelta(months=3)
    if range_from == '9m':
        start_date = end_date - relativedelta(months=9)
    if range_from == '1y':
        start_date = end_date - relativedelta(years=1)
    if range_from == '5y':
        start_date = end_date - relativedelta(years=5)
    if range_from == '10y':
        start_date = end_date - relativedelta(years=10)
    
    return str(start_date), str(end_date)


def get_range_timestamps(start_date, end_date):

    start_date = str(dt.strptime(start_date, "%Y-%m-%d")
        .timestamp()) \
        .replace('.0', '')

    end_date = str(dt.strptime(end_date, "%Y-%m-%d")
        .timestamp()) \
        .replace('.0', '')

    return start_date, end_date


def get_portfolio(assets, from_='3m', start_date=None, end_date=dt.now().date()):

    portfolio = pd.DataFrame()
    not_found = []
    for i, asset in tqdm(enumerate(assets), total = len(assets)):
        ticker_data = get_one_ticker(asset, from_=from_, start_date=start_date, end_date=end_date)
        if ticker_data is None:
            not_found.append(asset)
            continue
        close_data = prepare_data(ticker_data, asset).reset_index()
        if i == 0:
            portfolio = close_data[['Date', asset]]
            continue
        elif asset in portfolio.columns:
            continue
        portfolio = pd.concat([portfolio, close_data[asset]], axis=1)
        sleep(.7)

    portfolio = portfolio.set_index('Date')
    portfolio.index = pd.to_datetime(portfolio.index)
    portfolio = portfolio.round(3)
    print(f"Not found assets: {len(set(not_found))}, {set(not_found)}")
    return portfolio


def get_one_ticker(asset, from_='3m', start_date=None, end_date=dt.now().date()):
    
    headers = {
        'User-Agent': "Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.2; .NET CLR 1.0.3705;)",
    }

    if start_date:
        end_date = str(end_date)
        start, end = get_range_timestamps(start_date, end_date)
    else:
        start, end = get_range(range_from=from_, end_date=end_date)
        start, end = get_range_timestamps(start, end)

    url = f"https://query2.finance.yahoo.com/v8/finance/chart/{asset}?" \
          f"period1={start}&period2={end}&interval=1d&events=history"
    
    try:
        response = requests.get(url, headers=headers)
    except Exception as e:
        raise 
    if response.ok:
        data = pd.DataFrame.from_dict(response.json()['chart']['result'][0]['indicators']['quote'][0])
        data['Date'] = response.json()['chart']['result'][0]['timestamp']
        data['Date'] = data['Date'].apply(dt.fromtimestamp)
        data = data.set_index('Date', drop=True)
        data.index = data.index.normalize()
        return data.round(2)
    else:
        err = Exception(f'Could not retrieve data for {asset}:{response.text}')
        return None

    
def prepare_data(data, ticker):
    data = data.drop(['open', 'high', 'low', 'volume'], axis=1)
    data = data.rename(columns={"close": ticker})
    return data.round(2)



import pandas as pd
from earningspy.generators.finviz.data import (
    get_by_earnings_date, 
    get_by_tickers, 
    get_filters,
)

from earningspy.calendars.utils import calendar_pre_formatter, days_left
from earningspy.common.constants import (  
    FINVIZ_EARNINGS_DATE_KEY,
    DAYS_LEFT_KEY,
)

class EarningSpy:

    @classmethod
    def filters(cls, *args, **kwargs):
        return get_filters(*args, **kwargs)

    @classmethod
    def get_calendar(cls, sector=None, industry=None, index=None, future_only=True):
        
        finviz_data = cls.get_finviz(sector=sector, 
                                     industry=industry, 
                                     index=index)

        finviz_data = cls._arrange(finviz_data)
        if future_only:
            finviz_data = finviz_data[finviz_data[DAYS_LEFT_KEY] >= 0]
        return finviz_data


    @classmethod
    def get_this_week_earnings(cls):
        data = get_by_earnings_date(scope="this_week")
        return cls._arrange(data)

    @classmethod
    def get_previous_week_earnings(cls):
        data = get_by_earnings_date(scope="last_week")
        return cls._arrange(data)

    @classmethod
    def get_next_week_earnings(cls):
        data = get_by_earnings_date(scope="next_week")
        return cls._arrange(data)

    @classmethod
    def get_this_month_earnings(cls):
        data = get_by_earnings_date(scope="this_month")
        return cls._arrange(data)

    @classmethod
    def get_today_bmo(cls):
        data = get_by_earnings_date(scope="today_bmo")
        return cls._arrange(data)

    @classmethod
    def get_yesterday_amc(cls):
        data = get_by_earnings_date(scope="yesterday_amc")
        return cls._arrange(data)
    
    @classmethod
    def get_by_tickers(cls, tickers):
        """
        Get earnings data for a list of tickers.
        """
        if not isinstance(tickers, list):
            raise ValueError("tickers must be a list of strings")
        
        data = get_by_tickers(tickers)
        return cls._arrange(data)

    @classmethod
    def _check_missing_dates(cls, finviz_data):
        missing_count = finviz_data.index.to_series().apply(
            lambda row: isinstance(row, pd._libs.tslibs.nattype.NaTType) or pd.isna(row) or row is None
        ).sum()
        if missing_count > 0:
            print(f"[WARNING] Found {missing_count} missing earnings dates in the data.")

    @classmethod
    def _compute_days_left(cls, finviz_data):
        
        cls._check_missing_dates(finviz_data)
        finviz_data[DAYS_LEFT_KEY] = finviz_data.apply(lambda row: days_left(row), axis=1)
        return finviz_data

    @classmethod
    def _arrange(cls, data):
        data = data.set_index(FINVIZ_EARNINGS_DATE_KEY, drop=True)
        data = data.sort_index(ascending=True)
        data = cls._compute_days_left(data)
        return calendar_pre_formatter(data)



import asyncio
import os
from typing import Callable, Dict, List

import aiohttp
import requests
import tenacity
import urllib3
from lxml import html
from requests import Response
from tqdm import tqdm
from user_agent import generate_user_agent

from earningspy.generators.finviz.config import connection_settings
from earningspy.generators.finviz.helper_functions.error_handling import ConnectionTimeout

urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)


def http_request_get(
    url, session=None, payload=None, parse=True, user_agent=generate_user_agent()
):
    """ Sends a GET HTTP request to a website and returns its HTML content and full url address. """

    if payload is None:
        payload = {}

    try:
        if session:
            content = session.get(
                url,
                params=payload,
                verify=False,
                headers={"User-Agent": user_agent},
            )
        else:
            content = requests.get(
                url,
                params=payload,
                verify=False,
                headers={"User-Agent": user_agent},
            )

        content.raise_for_status()  # Raise HTTPError for bad requests (4xx or 5xx)
        if parse:
            return html.fromstring(content.text), content.url
        else:
            return content.text, content.url
    except (asyncio.TimeoutError, requests.exceptions.Timeout):
        raise ConnectionTimeout(url)


@tenacity.retry(wait=tenacity.wait_exponential())
def finviz_request(url: str, user_agent: str) -> Response:
    response = requests.get(url, headers={"User-Agent": user_agent})
    if response.text == "Too many requests.":
        raise Exception("Too many requests.")
    return response


def sequential_data_scrape(
    scrape_func: Callable, urls: List[str], user_agent: str, *args, **kwargs
) -> List[Dict]:
    data = []

    for url in tqdm(urls, disable="DISABLE_TQDM" in os.environ):
        try:
            response = finviz_request(url, user_agent)
            kwargs["URL"] = url
            data.append(scrape_func(response, *args, **kwargs))
        except Exception as exc:
            raise exc

    return data


class Connector:
    """ Used to make asynchronous HTTP requests. """

    def __init__(
        self,
        scrape_function: Callable,
        urls: List[str],
        user_agent: str,
        *args,
        css_select: bool = False
    ):
        self.scrape_function = scrape_function
        self.urls = urls
        self.user_agent = user_agent
        self.arguments = args
        self.css_select = css_select
        self.data = []

    async def __http_request__async(
        self,
        url: str,
        session: aiohttp.ClientSession,
    ):
        """ Sends asynchronous http request to URL address and scrapes the webpage. """

        try:
            async with session.get(
                url, headers={"User-Agent": self.user_agent}
            ) as response:
                page_html = await response.read()

                if page_html.decode("utf-8") == "Too many requests.":
                    raise Exception("Too many requests.")

                if self.css_select:
                    return self.scrape_function(
                        html.fromstring(page_html), *self.arguments
                    )
                return self.scrape_function(page_html, *self.arguments)
        except (asyncio.TimeoutError, requests.exceptions.Timeout):
            raise ConnectionTimeout(url)

    async def __async_scraper(self):
        """ Adds a URL's into a list of tasks and requests their response asynchronously. """

        async_tasks = []
        conn = aiohttp.TCPConnector(
            limit_per_host=connection_settings["CONCURRENT_CONNECTIONS"]
        )
        timeout = aiohttp.ClientTimeout(total=connection_settings["CONNECTION_TIMEOUT"])

        async with aiohttp.ClientSession(
            connector=conn, timeout=timeout, headers={"User-Agent": self.user_agent}
        ) as session:
            for url in self.urls:
                async_tasks.append(self.__http_request__async(url, session))

            self.data = await asyncio.gather(*async_tasks)

    def run_connector(self):
        """ Starts the asynchronous loop and returns the scraped data. """

        asyncio.set_event_loop(asyncio.SelectorEventLoop())
        loop = asyncio.get_event_loop()
        loop.run_until_complete(self.__async_scraper())

        return self.data



from datetime import datetime
import pandas as pd
from earningspy.common.constants import (
    FIELDS_ORDER
)


def calendar_pre_formatter(data):

    data.columns = data.columns.str.replace(' ', '_')
    data.columns = data.columns.str.upper()
    data.columns = data.columns.str.strip()

    data = data.convert_dtypes().infer_objects()

    data = data.round(4)
    data = data[FIELDS_ORDER]
    return data


def days_left(row):
    if isinstance(row.name, pd._libs.tslibs.nattype.NaTType) or pd.isna(row.name) or row.name is None:
        return None
    
    earnings_date = row.name

    if hasattr(earnings_date, "date"):
        earnings_date = earnings_date.date()

    today = datetime.today().date()
    diff = (earnings_date - today).days
    return diff
